/*
 * (C) Copyright 1996- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation nor
 * does it submit to any jurisdiction.
 */

/// @file   Reindexer.h
/// @author Chris Bradley

#pragma once

#include "fdb5/database/Archiver.h"
#include "fdb5/database/FieldLocation.h"

namespace fdb5 {

//----------------------------------------------------------------------------------------------------------------------
// Used for writing fields into a catalogue without writin any data to a store.
class Reindexer : public Archiver {

public:  // methods

    Reindexer(const Config& dbConfig = Config().expandConfig());

    virtual ~Reindexer();

    // Write a field location to the catalogue without touching the store
    void reindex(const Key& key, const FieldLocation& fieldLocation);

protected:  // methods

    void flushDatabase(Database& db) override;
};

//----------------------------------------------------------------------------------------------------------------------

}  // namespace fdb5
